//-------------------------------------------------------------------------------------------------
/// \file	BasicHwShader.cpp
/// \date	6-11-04
/// \author	Rob Bateman
/// \brief	This impliments possibly one of the most basic Hardware shaders available. It simply
///			aims to recreate the basic OpenGL material properties, ie ambient, diffuse, specular
/// 		emission and shininess.
/// \note	To use this shader, you must connect the outColor attribute of the shader to the
/// 		hardwareShader input on a lambert, phong, blinn, phongE or anisotropic material. You
/// 		must also ensure that you are viewing the shader in textured mode, ie press 6 when
/// 		in the viewport.
///
/// \todo	Before this plugin is fully ready for production use, we still have to handle the
/// 		following things:
///
///			1. The AE template to get the attributes to correctly display in the Attribute Editor
/// 		2. We could do with an icon for the shader, just a little final bit of polish you
/// 		   understand....
/// 		3. We need to provide an MPxDragAndDropBehaviour class so that maya knows how to connect
/// 		   this node to other nodes & attributes.
/// 		4. Textures as yet are unsupported. ie, if you connect the color attribute of the
/// 		   hardware shader to a file texture, we will not get the texturing as yet...
///
//-------------------------------------------------------------------------------------------------

#include "BasicSwShader.h"
#include <maya/MFnLightDataAttribute.h>
#include <maya/MRenderUtil.h>
#include <fstream>

#define MAKE_INPUT(attr)	\
	attr.setKeyable(true);	\
	attr.setStorable(true);	\
	attr.setReadable(true);	\
	attr.setWritable(true);

#define MAKE_OUTPUT(attr)		\
	attr.setKeyable(false); 	\
	attr.setStorable(false);	\
	attr.setReadable(true); 	\
	attr.setWritable(false);

const MTypeId BasicSwShader::m_TypeId( 0x70033 );
const MString BasicSwShader::m_TypeName( "basicSwShader" );


/// The diffuse colour of the material
MObject  BasicSwShader::m_aDiffuseColor;

/// The ambient colour of the material
MObject  BasicSwShader::m_aAmbientColor;

/// The diffuse colour of the material
MObject  BasicSwShader::m_aDiffuseReflectivity;

/// The output colour of the material
MObject  BasicSwShader::m_aOutColor;

/// the camera position
MObject  BasicSwShader::m_aPointCamera;

/// the surface normal
MObject  BasicSwShader::m_aNormalCamera;

/// the light direction
MObject  BasicSwShader::m_aLightDirection;

/// the lights intensity
MObject  BasicSwShader::m_aLightIntensity;

// three booleans used specifically for render layers. Maya will set these to true or
// false to tell us whether we should calculate the various components.
MObject  BasicSwShader::m_aLightAmbient;
MObject  BasicSwShader::m_aLightDiffuse;
MObject  BasicSwShader::m_aLightSpecular;
MObject  BasicSwShader::m_aLightShadowFraction;
MObject  BasicSwShader::m_aPreShadowIntensity;

// additional lighting info
MObject  BasicSwShader::m_aLightBlindData;
MObject  BasicSwShader::m_aLightData;


//
void* BasicSwShader::creator()
{
    return new BasicSwShader();
}

//
MStatus BasicSwShader::initialize()
{
	std::ofstream ofs("/cgstaff/rbateman/test.txt");

	// we will use a numeric attribute function set to add all of
	// the attributes to the node description.
    MFnNumericAttribute nAttr;
    MFnLightDataAttribute lAttr;


    // create the diffuse colour attribute
	m_aDiffuseColor =  nAttr.createColor( "color", "c");
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.7f, 0.7f, 0.7f);

    // create the ambient colour attribute
	m_aAmbientColor =  nAttr.createColor( "ambientColor", "ac");
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.4f, 0.4f, 0.4f);

	// a diffuse reflectivity value
	m_aDiffuseReflectivity = nAttr.create("diffuseReflectivity", "drfl", MFnNumericData::kFloat);
	MAKE_INPUT(nAttr);
	nAttr.setDefault(0.7f);

    // create the output colour attribute
	m_aOutColor =  nAttr.createColor( "outColor", "oc");
	MAKE_OUTPUT(nAttr);

	// the camera's position
    m_aPointCamera = nAttr.createPoint( "pointCamera", "pc" );
    MAKE_INPUT(nAttr);
    nAttr.setDefault(1.0f, 1.0f, 1.0f);
    nAttr.setHidden(true);

	// the input surface normal
    m_aNormalCamera = nAttr.createPoint( "normalCamera", "n" );
    MAKE_INPUT(nAttr);
    nAttr.setDefault(1.0f, 1.0f, 1.0f);
    nAttr.setHidden(true);

	//--------------------------------------------------------------------------------------
	//	The following attributes will be added as a compound array attribute, lightDataArray.
	//
	//	Maya will fill this array with info about each light in the scene. We can then use
	//	that array within compute to calculate the lights value. In this example we shall
	//	keep it simple and just calculate the diffuse lighting....
	//

		// the direction of a light
		m_aLightDirection = nAttr.createPoint( "lightDirection", "ld" );
		MAKE_INPUT(nAttr);
		nAttr.setDefault(1.0f, 1.0f, 1.0f);
		nAttr.setHidden(true);

		// a lights intensity
		m_aLightIntensity = nAttr.createColor( "lightIntensity", "li" );
		MAKE_INPUT(nAttr);
		CHECK_MSTATUS ( nAttr.setDefault(1.0f, 1.0f, 1.0f) );
		nAttr.setHidden(true);

		// a flag to indicate that the ambient colour should be calculated
		m_aLightAmbient = nAttr.create( "lightAmbient", "la", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// a flag to indicate that the diffuse colour should be calculated
		m_aLightDiffuse = nAttr.create( "lightDiffuse", "ldf", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// a flag to indicate that the specular colour should be calculated
		m_aLightSpecular = nAttr.create( "lightSpecular", "ls", MFnNumericData::kBoolean);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);
		nAttr.setDefault(true);

		// other data for a camera
		m_aLightBlindData = nAttr.create("lightBlindData", "lbld", MFnNumericData::kLong);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

		// the lights shadow value
		m_aLightShadowFraction = nAttr.create("lightShadowFraction", "lsf", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

		// intensity before the shadow
		m_aPreShadowIntensity = nAttr.create("preShadowIntensity", "psi", MFnNumericData::kFloat);
		MAKE_INPUT(nAttr);
		nAttr.setHidden(true);

	//--------------------------------------------------------------------------------------

	// create a compound light array attribute
    m_aLightData = lAttr.create( "lightDataArray", "ltd",
                               m_aLightDirection,
							   m_aLightIntensity,
							   m_aLightAmbient,
                               m_aLightDiffuse,
							   m_aLightSpecular,
							   m_aLightShadowFraction,
                               m_aPreShadowIntensity,
                               m_aLightBlindData);
    lAttr.setArray(true);
    lAttr.setStorable(false);
    lAttr.setHidden(true);
    lAttr.setDefault(0.0f, 0.0f, 0.0f, 0.0f, 0.0f, 0.0f, true, true, false, 0.0f, 1.0f, 0);

	// add the various attributes to the node
	addAttribute(m_aAmbientColor);
	addAttribute(m_aDiffuseColor);
	addAttribute(m_aDiffuseReflectivity);
	addAttribute(m_aPointCamera);
	addAttribute(m_aNormalCamera);
	addAttribute(m_aLightData);
	addAttribute(m_aOutColor);

	// set the attribute dependencies
	attributeAffects(m_aAmbientColor,  m_aOutColor);
	attributeAffects(m_aDiffuseColor,  m_aOutColor);
	attributeAffects(m_aDiffuseReflectivity,  m_aOutColor);
	attributeAffects(m_aPointCamera,  m_aOutColor);
	attributeAffects(m_aNormalCamera,  m_aOutColor);
	attributeAffects(m_aLightDirection,  m_aOutColor);
	attributeAffects(m_aLightIntensity,  m_aOutColor);
	attributeAffects(m_aLightAmbient,  m_aOutColor);
	attributeAffects(m_aLightDiffuse,  m_aOutColor);
	attributeAffects(m_aLightSpecular,  m_aOutColor);
	attributeAffects(m_aLightShadowFraction,  m_aOutColor);
	attributeAffects(m_aPreShadowIntensity,  m_aOutColor);
	attributeAffects(m_aLightBlindData,  m_aOutColor);

	// done
    return MS::kSuccess;
}


// computes the output colour. In a software shader this would be used to calculate an output
// colour for a pixel during rendering. There are 2 occasions when this takes place...
//
// 1. When the scene is rendered
// 2. When the sample swatch in the Attribute Editor gets rendered.
//
// In our case then, it will only get called when the swatch gets evaluated. Only software shaders
// can get rendered by the maya renderer...
//
MStatus BasicSwShader::compute (const MPlug& plug, MDataBlock& data)
{
	// Check that the requested recompute is one of the output values
	//
	if( (plug != m_aOutColor) &&
		(plug.parent() != m_aOutColor) )
		return MS::kUnknownParameter;

	// This will store the calculated diffuse lighting term
	double ambientTerm[3] = {0,0,0};

	// This will store the calculated diffuse lighting term
	double diffuseTerm[3] = {0,0,0};

    // get sample surface shading parameters
    MFloatVector& surfaceNormal = data.inputValue( m_aNormalCamera ).asFloatVector();
    MFloatVector& cameraPosition = data.inputValue( m_aPointCamera ).asFloatVector();

	// we will use the diffuse reflectivity as a way of globally controlling the intensity
	// of the diffuse colour. This is use below in the diffuse calculation
	float diffuseReflectivity = data.inputValue(m_aDiffuseColor).asFloat();

	// used for raytracing api enhancement below
	MFloatVector normal = surfaceNormal;
	MFloatVector point = cameraPosition;

    // get the array of lights
    MArrayDataHandle lightData = data.inputArrayValue( m_aLightData );
    int numLights = lightData.elementCount();

 	// iterate through light list and accumulate the diffuse lighting values
	for( int count=1; count <= numLights; count++ )
	{
		// get a handle to the current light
		MDataHandle currentLight = lightData.inputValue();

		// get references to the light intensity and direction
		MFloatVector& lightIntensity = currentLight.child(m_aLightIntensity).asFloatVector();
		MFloatVector& lightDirection = currentLight.child(m_aLightDirection).asFloatVector();

		// calculate ambient component
		if( currentLight.child(m_aLightDiffuse).asBool() )
		{
			ambientTerm[0] += lightIntensity[0];
			ambientTerm[1] += lightIntensity[1];
			ambientTerm[2] += lightIntensity[2];
		}

		// Find the blind data for the light
		int& blindData = currentLight.child( m_aLightBlindData ).asInt();
		if ( blindData == 0 )
		{

			// if diffuse is to be calculated, add in diffuse term
			if( currentLight.child(m_aLightDiffuse).asBool() )
			{
				// dot product the light direction and surface normal, L.N
				float LN = lightDirection * surfaceNormal;

				// calculate only if facing light
				if( LN > 0.0f )
				{
					diffuseTerm[0] += lightIntensity[0] * LN * diffuseReflectivity;
					diffuseTerm[1] += lightIntensity[1] * LN * diffuseReflectivity;
					diffuseTerm[2] += lightIntensity[2] * LN * diffuseReflectivity;
			    }
			}
        }
        else
        {
			// calculate LN using reflected rays
			float LN = MRenderUtil::diffuseReflectance( blindData, lightDirection, point, surfaceNormal, true );

			// calculate only if facing light
			if( LN > 0.0f )
			{
			     diffuseTerm[0] += lightIntensity[0] * LN * diffuseReflectivity;
			     diffuseTerm[1] += lightIntensity[1] * LN * diffuseReflectivity;
			     diffuseTerm[2] += lightIntensity[2] * LN * diffuseReflectivity;
			}
		}

		// next light please
		if( !lightData.next() )
			break;
	}

	// use the handle to get the diffuse color value
	const MFloatVector& color = data.inputValue(m_aDiffuseColor).asFloatVector();
	const MFloatVector& ambientColor = data.inputValue(m_aAmbientColor).asFloatVector();


	// again, get hold of a handle to the output attribute... This time
	// however we get an outputHandle.
	//
	MDataHandle outColorHandle = data.outputValue( m_aOutColor );

	// get a reference to the actual data
    MFloatVector & outColor = outColorHandle.asFloatVector();

	// set the output colour, diffuseTerm * diffuse colour
	outColor[0] = static_cast<float>(ambientColor[0] * ambientTerm[0] + color[0] * diffuseTerm[0]);
	outColor[1] = static_cast<float>(ambientColor[1] * ambientTerm[1] + color[1] * diffuseTerm[1]);
	outColor[2] = static_cast<float>(ambientColor[2] * ambientTerm[2] + color[2] * diffuseTerm[2]);

	// clean the plug, ie flag it as clean so that maya does not re-evaluate this attribute
	outColorHandle.setClean();

	return MS::kSuccess;
}
